        <uib-accordion close-others="oneAtATime">
            <div uib-accordion-group
                 class="panel-default"
                 heading="{{ 'Registration Code'|translate }}">
                {{ enrolledToken.registrationcode }}
            </div>
        </uib-accordion>

        <p class="help-block" translate>
            You may pass this registration code to the user.
        </p>
